import torch

class Permute(torch.nn.Module):
    def __init__(self, shape):
        super().__init__()
        self.shape = shape
        

    def forward(self, x):
        x = torch.permute(x, self.shape)
        return x



import torch

#Reshape((1, 28, 28))
#Reshape((28 * 28,))
#Reshape((28 * 28))
class Reshape(torch.nn.Module):
    def __init__(self, shape):
        super().__init__()
        self.shape = shape
        if type(self.shape) == int:
            self.shape = [self.shape]

    def forward(self, x):
        return x.view((-1, *self.shape))



import torch

#이전 시퀀스의 숨은 상태가 다음 시퀀스의 영행을 주기 않게 하기위해 새로운 시퀀스 입력시 이전 숨은 상태 초기화하는 LSTM
class HiddenStateResetLSTM(torch.nn.Module):
    def __init__(self, input_size, hidden_size, num_layers=1, **kwargs):
        super().__init__()
        self.layer = torch.nn.LSTM(input_size=input_size, hidden_size=hidden_size, num_layers=num_layers, **kwargs)
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers

    def forward(self, x, **kwargs):
        batch_length = len(x)
        hidden_state = torch.zeros(self.num_layers, batch_length, self.hidden_size)
        cell_state = torch.zeros(self.num_layers, batch_length, self.hidden_size)
        '''
        device = "cpu"
        if torch.cuda.is_available():
            device = "cuda"
        elif torch.backends.mps.is_available():
            device = "mps"
        '''
        hidden_state = hidden_state.to(x.device)
        cell_state = cell_state.to(x.device)
        x = self.layer(x, hx=(hidden_state, cell_state), **kwargs)
        return x



import torch

#https://en.wikipedia.org/wiki/Dot_product
class DotProduct(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, user_id, item_id):
        #print(user_id.shape) #torch.Size([1, 1, 100])
        #print(user_id.shape) #torch.Size([1, 1, 100])
        dot_product = (user_id * item_id).sum(2)
        #print(dot_product.shape) #torch.Size([1, 1])
        return dot_product



import torch

class LazilyInitializedLinear(torch.nn.Module):
    def __init__(self, out_features):
        super().__init__()
        self.layer = None
        self.out_features = out_features

    def forward(self, x):
        if not self.layer:
            in_features = x.shape[1]
            self.layer = torch.nn.Linear(in_features=in_features, out_features=self.out_features)
            '''
            device = "cpu"
            if torch.cuda.is_available():
                device = "cuda"
            elif torch.backends.mps.is_available():
                device = "mps"
            self.layer = self.layer.to(device)
            '''
            self.layer = self.layer.to(x.device)
        x = self.layer(x)
        return x



from . import dict_to_parameters
from . import dot_product
from . import gru_last_hidden_state
from . import hidden_state_reset_gru
from . import hidden_state_reset_lstm
from . import hidden_state_reset_rnn
from . import lazily_initialized_linear
from . import lstm_last_hidden_state
from . import reshape
from . import rnn_last_hidden_state
from . import select_from_array
from . import permute

from .dict_to_parameters import DictToParameters
from .dot_product import DotProduct
from .gru_last_hidden_state import GRULastHiddenState
from .hidden_state_reset_gru import HiddenStateResetGRU
from .hidden_state_reset_lstm import HiddenStateResetLSTM
from .hidden_state_reset_rnn import HiddenStateResetRNN
from .lazily_initialized_linear import LazilyInitializedLinear
from .lstm_last_hidden_state import LSTMLastHiddenState
from .reshape import Reshape
from .rnn_last_hidden_state import RNNLastHiddenState
from .select_from_array import SelectFromArray
from .permute import Permute



from bs4 import BeautifulSoup
import re

#print(clean_english('테스트 입니다 test this is <b>')) #테스트 입니다 test this is 
def clean_english(text):
    text = BeautifulSoup(text, 'html.parser').text #HTML 태그 제거
    text = text.lower() # 소문자화
    text = re.sub('[!"#$%&\()*+,-./:;<=>?@[\\]^_`{|}~]', repl= ' ', string=text) #특수기호 제거, 정규 표현식
    return text

import re
from bs4 import BeautifulSoup

#print(clean_korean('테스트 입니다 test this is <b>')) #테스트 입니다
def clean_korean(text):
    #텍스트 정제 (HTML 태그 제거)
    text = BeautifulSoup(text, 'html.parser').text 
    #텍스트 정제 (특수기호 제거)
    text = re.sub(r'[^ ㄱ-ㅣ가-힣]', '', text) #특수기호 제거, 정규 표현식
    return text



import numpy as np

#window_length = sequence_length + 1
def slice_time_series_data_from_np_array(np_array, x_column_indexes=None, label_column_indexes=None, sequence_length=7):
    #print(np_array.shape) #(980, 1)
    window_length = sequence_length + 1
    x = []
    labels = []
    for i in range(0, len(np_array) - window_length + 1): #0 ~ (980 - 4 - 1) 
        window = np_array[i:i + window_length, :]
        if x_column_indexes:
            x.append(window[:-1, x_column_indexes])
        else:
            labels.append(window[:-1, :])
        if label_column_indexes:
            labels.append(window[-1, label_column_indexes])
        else:
            labels.append(window[-1, :])
    x = np.array(x)
    labels = np.array(labels)
    #print(x.shape) #(977, 3, 1)
    #print(labels.shape) #(977, 1)
    return x, labels 



import torch

class RNNLastHiddenState(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        output = x[0] #모든 타임 스텝(토큰: 숫자, 문자, 단어)의 숨은 상태, 
        hidden_state = x[1] #마지막 타임 스텝(토큰: 숫자, 문자, 단어)의 숨은 상태
        #print(output.shape) #torch.Size([8, 380, 32]) 
        #print(hidden_state.shape) #torch.Size([1, 8, 32]) #레이어 수, 타임 스텝, 숨은 상태
        x = output[:,-1]
        #print(x)
        x = hidden_state[-1]
        #print(x)
        #print(x.shape) #torch.Size([8, 32])
        return x
